import discord
from cpl_core.logging import LoggerABC
from cpl_core.configuration import ConfigurationABC
from gismo_core.abc.bot_service_abc import BotServiceABC
from modules.permission.abc.permission_service_abc import PermissionServiceABC
from modules.permission.configuration.permission_settings import PermissionSettings


class PermissionService(PermissionServiceABC):

    def __init__(self, logger: LoggerABC, bot: BotServiceABC, config: ConfigurationABC):
        PermissionServiceABC.__init__(self)
        self._logger = logger
        self._bot = bot
        self._config = config

        self._admin_role_ids: dict[str, list[int]] = {}
        self._admin_roles: dict[list[discord.Role]] = {}
        self._admins: dict[list[discord.Member]] = {}
        
        self._moderator_role_ids: dict[list[int]] = {}
        self._moderator_roles: dict[list[discord.Role]] = {}
        self._moderators: dict[list[discord.Member]] = {}

    def on_ready(self): 
        for guild in self._bot.guilds:
            guild: discord.Guild = guild
            self._logger.debug(__name__, f'Validate permission settings')

            settings: PermissionSettings = self._config.get_configuration(f'Permission_{guild.id}')
            if settings is None:
                self._logger.error(__name__, 'Permission settings not found')
                return
            
            self._admin_role_ids[guild.id] = settings.admin_roles
            self._moderator_role_ids[guild.id] = settings.moderator_roles

            admin_roles = []
            admins = []

            mod_roles = []
            mods = []

            for role in guild.roles:
                role: discord.Role = role

                if role.id in self._admin_role_ids[guild.id]:
                    admin_roles.append(role)
                    self._logger.trace(__name__, f'Added admin role {role}')

                    for member in role.members:
                        admins.append(member)
                        self._logger.trace(__name__, f'Added admin {member}')

                if role.id in self._moderator_role_ids[guild.id]:
                    mod_roles.append(role)
                    self._logger.trace(__name__, f'Added moderator role {role}')

                    for member in role.members:
                        mods.append(member)
                        self._logger.trace(__name__, f'Added moderator {member}')
            

            self._admin_roles[guild.id] = admin_roles
            self._admins[guild.id] = admins
            self._moderator_roles[guild.id] = mod_roles
            self._moderators[guild.id] = mods

    def on_member_update(self, before: discord.Member, after: discord.Member):
        g_id = after.guild.id

        for admin_role in self._admin_roles[g_id]:
            if admin_role in before.roles  and admin_role not in after.roles:
                self._admins[g_id].remove(after)
                self._logger.trace(__name__, f'Removed {after.id} from admins')

            elif admin_role in after.roles  and admin_role not in before.roles:
                self._admins[g_id].append(after)
                self._logger.trace(__name__, f'Added {after.id} to admins')
        
        for moderator_role in self._moderator_roles[g_id]:
            if moderator_role in before.roles  and moderator_role not in after.roles:
                self._moderators[g_id].remove(after)
                self._logger.trace(__name__, f'Removed {after.id} from moderators')

            elif moderator_role in after.roles  and moderator_role not in before.roles:
                self._moderators[g_id].append(after)
                self._logger.trace(__name__, f'Added {after.id} to moderators')


    def get_admin_role_ids(self, g_id: int) -> list[int]:
        return self._admin_role_ids[g_id]

    def get_admin_roles(self, g_id: int) -> list[discord.Role]:
        return self._admin_roles[g_id]

    def get_admins(self, g_id: int) -> list[discord.Member]:
        return self._admins[g_id]

    def get_moderator_role_ids(self, g_id: int) -> list[int]:
        return self._moderator_role_ids[g_id]

    def get_moderator_roles(self, g_id: int) -> list[discord.Role]:
        return self._moderator_roles[g_id]

    def get_moderators(self, g_id: int) -> list[discord.Member]:
        return self._moderators[g_id]
